from cellsaw.merge import Merge, stringdiffuse, accuracy_evaluation
from collections import Counter
import scanpy as sc
from cellsaw.annotate.draw import plot_annopair
from cellsaw.annotate.annotators import label_knn, linsum_copylabel, raw_diffusion, markercount, mergewrap, raw_diffusion_combat, tunnelclust, scanorama_integrate_diffusion
import cellsaw.preprocess as preprocess
import numpy as np


def predict_celltype(target,
                   source,
                   source_label = 'celltype',
                   target_label='predicted_celltype',
                   pca_dim = 20,
                   umap_dim = 0,
                   make_even = True,
                   sigmafac = 1,
                   n_intra_neighbors = 7,
                   n_inter_neighbors = 2,
                   premerged = False,
                   n_genes = 800,
                   pp = False,
                   linear_assignment_factor= 1,
                   similarity_scale_factor = 1.0):

    assert similarity_scale_factor == 1.0, 'not implemented'

    pid = (pca_dim>0)+ (umap_dim>0)
    #merged =premerged or  mergewrap(target,source,umap_dim,pca = pca_dim, make_even=make_even)
    if pp and not premerged:
        def prep(x):
            # x.X = np.expm1(x.X)
            # x._uns.pop("log1p")
            # x.uns
            preprocess.annotate_genescore_single(x,selector = pp)
            # sc.pp.log1p(x)
        prep(target)
        prep(source)
    merged = premerged or mergewrap(target,
                                   source,
                                   umap_dim,
                                   pca = pca_dim,
                                    selectgenes = n_genes,
                                   make_even=make_even, sortfield = -1)
    newlabels = stringdiffuse(merged,merged.data[1].obs[source_label],sigmafac=sigmafac,
            pid = pid,
            neighbors_inter=n_inter_neighbors,
            neighbors_intra=n_intra_neighbors, linear_assignment_factor=linear_assignment_factor)
    target.obs[target_label] = newlabels
    return merged
    #return target


def multi_annotate(target,
                   sources,
                   annotator = lambda x:x,
                   source_label = 'celltype',
                   target_label='multisrc',
                   annotatorargs = None):

    # we should not be loosing cells:
    sc.pp.subsample(target,
            fraction=None,
            n_obs=min([a.shape[0] for a in sources+[target]]),
            random_state=1227, copy=False)


    # annotate
    allobs = []
    for i,source in enumerate(sources):
        # we send a copy in as merge will delete genes :)
        target_tmp = annotator(target.copy(),source.copy(),
                source_label = source_label,
                target_label = f'{target_label}_{i}',
                **annotatorargs)
        allobs.append(target_tmp.obs[f'{target_label}_{i}'])

    def calclabel(a):
        c = Counter(a)
        maxcount = max(c.values())
        # so we return the majority,
        # problem might be that the cardinality might show up twice
        # this is no problem since counter seems to list keys in order of reading
        # so we get the 'left' most value in case of a collision
        for k,v in c.items():
            if maxcount ==v:
                return k

    target.obs[target_label] = [ calclabel(a) for a in zip(*allobs)  ]
    return target


def EXPERIMENTAL(target,
                   source,
                   source_label = 'celltype',
                   target_label='predicted_celltype',
                   pca_dim = 20,
                   umap_dim = 0,
                   make_even = True,
                   sigmafac = 1,
                   n_intra_neighbors = 7,
                   n_inter_neighbors = 2,
                   premerged = False,
                   n_genes = 800,
                   pp = False,
                   linear_assignment_factor= 1,
                   similarity_scale_factor = 1.0):
    import anndata as ad
    '''
    # i am experimenting with how the genes are selected..
    '''
    assert similarity_scale_factor == 1.0, 'not implemented'
    pid = (pca_dim>0)+ (umap_dim>0)
    #merged =premerged or  mergewrap(target,source,umap_dim,pca = pca_dim, make_even=make_even)


    #zz = ad.concat([target,source], join = 'outer')

    if False:
        from scipy.sparse import vstack
        zz = vstack([target.X,source.X])
        zz = ad.AnnData(X = zz)
        preprocess.annotate_genescore_single(zz,selector = pp)
        target.uns['lastscores']='hack'
        source.uns['lastscores']='hack'
        target.varm['hack'] = zz.varm[pp]
        source.varm['hack'] = zz.varm[pp]
    else:

        preprocess.annotate_genescore_single(target,selector = pp)
        preprocess.annotate_genescore_single(source,selector = pp)
        asd = np.vstack([target.varm[pp], source.varm[pp]])
        nuvalues = np.max(asd, axis = 0)

        #print(f"{ nuvalues.shape=} {asd.shape=}")
        target.uns['lastscores']='hack'
        source.uns['lastscores']='hack'
        target.varm['hack'] = nuvalues
        source.varm['hack'] = nuvalues



    merged =  mergewrap(target, source,
                                   umap_dim,
                                   pca = pca_dim,
                                    selectgenes = n_genes,
                                   make_even=make_even, sortfield = -1)

    newlabels = stringdiffuse(merged,merged.data[1].obs[source_label],sigmafac=sigmafac,
            pid = pid,
            neighbors_inter=n_inter_neighbors,
            neighbors_intra=n_intra_neighbors, linear_assignment_factor=linear_assignment_factor)
    target.obs[target_label] = newlabels
    return target





import numpy as np
import scanpy as sc
from cellsaw.merge.mergehelpers import hungarian
from sklearn.neighbors import KNeighborsClassifier as knn
from sklearn.semi_supervised import LabelPropagation as lapro
from sklearn.semi_supervised import LabelSpreading as laspre
from MarkerCount.marker_count import MarkerCount_Ref, MarkerCount
import warnings
from cellsaw.merge import Merge
from cellsaw.merge.hungarianEM import linasEM

def mergewrap(a,b,umap_dim,**kwargs):
    assert  isinstance(umap_dim, int), 'umap_dim must be an integer'
    umaps = [] if umap_dim == 0 else [umap_dim]
    m =  Merge([a, b], umaps=umaps, **kwargs)
    return m

def linsum_copylabel(
        target,
        source,
        source_label = 'celltype',
        target_label= 'diffuseknn',
        premerged = False,
        n_genes = 800,
        pca_dim = 20, umap_dim = 0):

    pid = (pca_dim>0)+ (umap_dim>0)
    merged =   premerged or mergewrap(target,source,umap_dim,pca=pca_dim, sortfield = pid, selectgenes = n_genes)
    target.obs[target_label] = list(merged.data[1].obs[source_label])
    return target


def label_knn(target,source,
                       source_label = 'celltype',
                       target_label='knn',
                        n_genes = 800,
                       premerged = False,
                       pca_dim = 20, umap_dim = 0, k = 5):


    pid = (pca_dim>0)+ (umap_dim>0)
    merged = premerged or  mergewrap(target,source,umap_dim,pca=pca_dim, selectgenes = n_genes)
    a,b = merged.projections[pid]
    y = merged.data[1].obs['celltype']
    model = knn(n_neighbors=k).fit(a,y)
    target.obs[target_label] = model.predict(b)
    return target


def raw_diffusion(target, source, source_label ='celltype',
                  target_label='raw_diffusion',
                  premerged = False,
                  n_neighbors = 5, gamma = 10,
                  n_genes = 800,
                  pca_dim = 20, umap_dim = 0):
    pid = (pca_dim>0)+ (umap_dim>0)
    merged = premerged or mergewrap(target,source,umap_dim,pca=pca_dim,selectgenes=n_genes)
    a,b = merged.projections[pid]
    y = merged.data[1].obs[source_label]
    #diffusor = laspre( gamma = .1, n_neighbors = 5, alpha = .4).fit(b,y)
    diffusor = lapro( gamma = gamma, n_neighbors = n_neighbors).fit(b,y)
    target.obs[target_label] = diffusor.predict(a)
    return target



def tunnelclust(target, source, source_label ='celltype',
                  target_label='raw_diffusion',
                  premerged = False,
                  n_neighbors = 5,
                  n_genes = 800,
                  pca_dim = 20,
                umap_neigh  = 10,
                    umap_dim = 0):
    '''
    change params,
    write multitunnelclust wrapper to work with strings
    change the code in this function to call that wrapper
    '''
    pid = (pca_dim>0)+ (umap_dim>0)
    merged = premerged or mergewrap(target,source,umap_dim,pca=pca_dim,selectgenes=n_genes)
    a,b = merged.projections[pid]
    y = merged.data[1].obs[source_label]
    # diffusor = laspre( gamma = .1, n_neighbors = 5, alpha = .4).fit(b,y)
    # diffusor = lapro( gamma = gamma, n_neighbors = n_neighbors).fit(b,y)
    target.obs[target_label] = linasEM([a,b],y)
    return target



def markercount(target, source, source_label ='celltype',
                target_label='markercount_celltype', premerged = False,
                pca_dim = 20, umap_dim = 0):


    pid = (pca_dim>0)+ (umap_dim>0)

    # merged = premerged or mergewrap(target,source,umap_dim,pca=pca_dim)
    # X_ref=merged.data[1].to_df()
    # X_test=merged.data[0].to_df()
    # reflabels = merged.data[1].obs[source_label]

    X_ref=source.to_df()
    X_test=target.to_df()
    reflabels = source.obs[source_label]

    #print(f"{X_ref.shape=}{X_test.shape=}{reflabels=}")
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        try:
            df_res = MarkerCount_Ref( X_ref, reflabels, X_test,
                                  cell_types_to_excl = ['Unknown'],
                                  log_transformed = True,
                                  file_to_save_marker = 'my_markers',
                                  verbose = False )
        except:
            # get results :D
            predict = ['failed to run markercount']*target.X.shape[0]
            target.obs[target_label] = predict
            return target
        # get results :D
        predict = df_res['cell_type_pred']
        target.obs[target_label] = predict
    return target


import anndata as ad
def raw_diffusion_combat(target, source, source_label ='celltype',
                  target_label='raw_diffusion',
                  premerged = False,
                  n_neighbors = 5, gamma = 10,
                  n_genes = 800,
                  pca_dim = 20, umap_dim = 0):
    pid = (pca_dim>0)+ (umap_dim>0)

    target.obs['batch'] = [1]*target.X.shape[0]
    source.obs['batch'] = [2]*source.X.shape[0]
    scoresvariable  = target.uns['lastscores']
    scr = [a.varm[scoresvariable] for a in [target,source]]
    z = ad.concat([target,source])
    z.obs_names_make_unique()
    sc.pp.combat(z,key= 'batch')
    target = z[z.obs['batch'] == 1]
    source = z[z.obs['batch'] == 2]
    #target.varm[scoresvariable]= scr[0]
    #source.varm[scoresvariable]= scr[1]
    #merged = premerged or mergewrap(target,source,umap_dim,pca=pca_dim,selectgenes= n_genes)
    #a,b = merged.projections[pid]
    '''
    todo dimension reduction pca
    dimred umap
    split a and b
    '''
    from sklearn.decomposition import PCA
    import umap
    x = PCA(n_components=pca_dim).fit_transform(z.X)
    if umap_dim > 0:
        x = umap.UMAP(n_components=umap_dim).fit_transform(x)

    a,b = x[z.obs['batch']==1], x[z.obs['batch']==2]

    y = source.obs[source_label]
    #diffusor = laspre( gamma = .1, n_neighbors = 5, alpha = .4).fit(b,y)
    diffusor = lapro( gamma = gamma, n_neighbors = n_neighbors).fit(b,y)
    target.obs[target_label] = diffusor.predict(a)
    return target


def scanorama_integrate_diffusion(target,source,
                       source_label = 'celltype',
                       target_label='scanorama',
                       pca_dim = 20, umap_dim = 0, gamma = .75):


    import scanorama
    from sklearn.decomposition import PCA
    import umap


    # make names unique and integrate
    target.obs['batch'] = [1]*target.X.shape[0]
    source.obs['batch'] = [2]*source.X.shape[0]
    z = ad.concat([target,source])
    z.obs_names_make_unique()
    a,b = z[z.obs['batch']==1], z[z.obs['batch']==2]
    scanorama.integrate_scanpy([a,b])

    # stack to do dimension reduction
    z = ad.concat([a,b])
    data = z.obsm['X_scanorama']
    x = PCA(n_components=pca_dim).fit_transform(data)
    if umap_dim > 0:
        x = umap.UMAP(n_components=umap_dim).fit_transform(x)

    # un-stack to do label transfer
    a,b = x[z.obs['batch']==1], x[z.obs['batch']==2]
    y = source.obs[source_label]
    #diffusor = laspre( gamma = .1, n_neighbors = 5, alpha = .4).fit(b,y)
    diffusor = lapro( gamma = gamma).fit(b,y)
    target.obs[target_label] = diffusor.predict(a)
    return target





import numpy as np
import scanpy as sc
import umap
from sklearn import decomposition
from sklearn.metrics import pairwise_distances
from scipy.optimize import linear_sum_assignment


def make_even(data):
        # assert all equal
        size = data[0].X.shape[1]
        assert all([size == other.X.shape[1] for other in data])

        # find smallest
        counts = [e.X.shape[0] for e in data]
        smallest = min(counts)

        for a in data:
            if a.X.shape[0] > smallest:
                sc.pp.subsample(a,
                                fraction=None,
                                n_obs=smallest,
                                random_state=0,
                                copy=False)
        return data


def unioncut(scores, numGenes, data):

    indices = np.argpartition(scores, -numGenes)[:,-numGenes:]
    indices = np.unique(indices.flatten())
    [d._inplace_subset_var(indices) for d in data]
    return data
    #return [d[:,indices].copy() for d in data]


def equal_contrib_gene_cut(scores,numgenes,data):
    ar = np.array(scores)
    ind = np.argsort(ar)

    def calcoverlap(scores,number):
        indices = scores[:,-number:]
        indices = np.unique(indices.flatten())
        return indices

    def findcutoff(low,high, lp = -1):
        probe = int((low+high)/2)
        y = calcoverlap(ind,probe)
        if probe == lp:
            return y
        if len(y) > numgenes:
            return findcutoff(low,probe,probe)
        else:
            return findcutoff(probe,high,probe)

    indices = findcutoff(0,numgenes)
    [d._inplace_subset_var(indices) for d in data]
    return data

def dimension_reduction(adatas, scale, zero_center, PCA, umaps, joint_space=True):
    # get a (scaled) dx
    if scale or PCA:
        adatas= [sc.pp.scale(adata, zero_center=False, copy=True,max_value=10) for adata in adatas]
    dx = [adata.to_df().to_numpy() for adata in adatas]
    if joint_space == False:
        return disjoint_dimension_reduction(dx, PCA, umaps)

    res = []
    if PCA:
        pca = decomposition.PCA(n_components=PCA)
        pca.fit(np.vstack(dx))
        #print('printing explained_variance\n',list(pca.explained_variance_ratio_))# rm this:)
        dx = [ pca.transform(e) for e in dx  ]
        res.append(dx)

    for dim in umaps:
        assert 0 < dim < PCA, f'{dim=} {PCA=}'
        res.append(umapify(dx,dim))

    return res


def umapify(dx, dimensions,n_neighbors=10):
    mymap = umap.UMAP(n_components=dimensions,
                      n_neighbors=n_neighbors,
                      random_state=1337).fit(np.vstack(dx))
    return [mymap.transform(a) for a in dx]


def disjoint_dimension_reduction(dx, PCA, umaps):
    res = []
    if PCA:
        pcaList = [decomposition.PCA(n_components=PCA) for x in dx]
        for pca, x in zip(pcaList, dx):
            pca.fit(np.vstack(x))
        #print('printing explained_variance\n',list(pca.explained_variance_ratio_))# rm this:)
        dx = [ pca.transform(e) for pca, e in zip(pcaList,dx)  ]
        res.append(dx)
    for dim in umaps:
        assert 0 < dim < PCA
        res.append(umapify(dx,dim))
    return res

def hungarian(X1, X2, debug = False,metric='euclidean'):
    # get the matches:
    # import time
    # print("STARTING CALC")
    # now = time.time()
    distances = pairwise_distances(X1,X2, metric=metric)
    #print(f" distcalc: {time.time() -now}")
    #distances = ed(X1, X2)

    #if solver != 'scipy':
    '''
    from time import time
    now = time()
    from lapjv import lapjv
    row_ind, col_ind, _ = lapjv(distances)
    print(f"  {time() - now}s"); now =time()
    from scipy.optimize import linear_sum_assignment
    row_ind, col_ind = linear_sum_assignment(distances)
    print(f"  {time() - now}s"); now =time()
    from lapsolver import solve_dense
    row_ind,col_ind = solve_dense(distances)
    print(f"  {time() - now}s"); now =time()
    '''
    from scipy.optimize import linear_sum_assignment
    row_ind, col_ind = linear_sum_assignment(distances)

    if debug:
        x = distances[row_ind, col_ind]
        num_bins = 100
        print("hungarian: debug hist")
        plt.hist(x, num_bins, facecolor='blue', alpha=0.5)
        plt.show()
    return (row_ind, col_ind), distances



from lmz import *
import scanpy as sc
import numpy as np




def rename_obs(datasets, batch, typ):
    for ds, bat, lab in zip(datasets, batch, typ):
        if batch != 'batch':
            ds.obs['batch'] = ds.obs[bat]
        if typ != 'label':
            ds.obs['label'] = ds.obs[lab]

def load_scib(path = '/home/ubuntu/benchdata/'):
    datasets = [sc.read(path+data) for data in "Immune_ALL_hum_mou.h5ad  Immune_ALL_human.h5ad  Lung_atlas_public.h5ad  human_pancreas_norm_complexBatch.h5ad".split()]
    batch,typ = Transpose (Map(lambda x:x.split(), 'batch final_annotation#batch final_annotation#batch cell_type#tech celltype'.split("#")))
    # rename fields
    rename_obs(datasets,batch, typ)
    # split by batch
    datasets =  [[z[z.obs['batch']==i] for i in z.obs['batch'].unique()] for z
                 in datasets]
    return datasets



def load_timeseries(path = './',remove_unassigned = True):
    datasets = [sc.read(path+data+".h5ad") for data in
                "s5 509 1290 mousecortex water pancreatic cerebellum".split()]
    if remove_unassigned:
        okcells =  lambda ds: [ l not in ['-1','Unassigned','nan'] for l in ds.obs['label']]
        datasets = [ds[okcells(ds)] for ds in datasets]

    datasets =  [[z[z.obs['batch']==i] for i in z.obs['batch'].unique()]
                 for z in datasets]
    return datasets


def test_ts():
    ds = load_timeseries()
    for i,adata in enumerate(ds):
        print(i)
        print(adata)
        for a in adata:
            print(np.unique(a.obs['label']))





from lmz import Map,Zip,Filter,Grouper,Range,Transpose, grouper
import numpy as np
from ubergauss import tools as ut
from sklearn.metrics.pairwise import cosine_similarity as cos
from sklearn import preprocessing
from sklearn.metrics import precision_score


def cosine(a,b, numgenes = 500, scores= 'scores'):
    scr1, scr2 = a.varm[scores], b.varm[scores]
    if numgenes:
        mask = scr1+scr2
        mask = ut.binarize(mask,numgenes).astype(np.bool)
        scr1 = scr1[mask]
        scr2 = scr2[mask]
    else:
        assert False
    return cos([scr1],[scr2]).item()



def jaccard(adata1,adata2 ,numgenes= False,scores='scores'):
    # intersect/union
    assert numgenes > 3
    asd = np.array([ ut.binarize(d.varm[scores],numgenes)  for d in [adata1,adata2]])
    union = np.sum(np.any(asd, axis=0))
    intersect = np.sum(np.sum(asd, axis=0) ==2)
    return intersect/union


def precision(matrix,shortlabels, k):
    truth = preprocessing.LabelEncoder().fit_transform(shortlabels)
    srt = np.argsort(matrix, axis=1)
    #print(f'{truth=} {srt=}')
    pred = truth[ [ srt[i,-j]  for j in range(1,k+1) for i in Range(truth)] ]
    true = np.hstack([truth for i in range(k)])
    return  precision_score(true, pred, average='micro')








def mkshortnames(li):
    return [l[:5] for l in li]

def precission_at(df,k):
    return precision(df.to_numpy(),mkshortnames(df.columns),k)







def matrixmap(inst,inst2,meth, scorelabel, genecount):
    l = len(inst)
    l2 = len(inst2)
    res = np.zeros((l,l2))
    f = jaccard if meth == 'jaccard' else cosine
    for i in range(l):
        for j in range(i,l2):
            a,b = inst[i], inst2[j]
            r= f(a,b,numgenes=genecount, scores= scorelabel)
            if r==1.0: # it is 1 if inst and inst2 are the same -> genescore arrays are identical
                r = -1
            res[i,j] = r
            res[j,i] = r
    res[res < 0] = res.max()*1.05
    return res

def adata_to_score(instances,genecount, preprocessing, cosjacc, labels):
    shortlabels = mkshortnames(labels)
    matrix = matrixmap(instances, instances,cosjacc,preprocessing, genecount)
    return [precision(matrix,shortlabels,k) for k in [1,2,3]]
    # make the matrix
    # score the k@k




import matplotlib.pyplot as plt
from cellsaw.merge import  Merge
from cellsaw import draw
from ubergauss import tools as ut
import numpy as np
import seaborn as sns

def plot_annopair(source,target,source_label = '', target_label ='', pca= 20):
    '''
    annotate will transfer labels between andata objects -> plot these
    '''

    if id(source) == id(target):
        datasets = [target, target.copy()]
    else:
        datasets = [source, target]


    umaps = [2] if pca >2 else []
    merged = Merge(datasets, umaps=umaps, pca = pca, make_even = False, sortfield=-1)

    s,t = merged.projections[2] if pca > 2 else merged.projections[1]


    sns.set_theme(style = 'whitegrid')
    d = draw.tinyUmap(dim=(1,3), lim = [s,t])

    tlab = list(merged.data[1].obs[target_label])
    slab = list(merged.data[0].obs[source_label])
    allthelabels = tlab + slab
    all,sm = ut.labelsToIntList(allthelabels)


    labeld = {k:v[:8] for k,v in sm.getitem.items()}

    size = 10
    d.draw(t,sm.encode(tlab), title = f'T:{target_label}', labeldict = labeld,size= size)

    def plot_cellids(t):
        [plt.text(a, b, str(i),fontsize='xx-small') for i,(a,b) in enumerate(t.tolist())]

    # plot_cellids(t)
    d.draw(s,sm.encode(slab), title = f'S:{source_label}', labeldict = labeld, size=size)
    # plot_cellids(s)
    d.draw(np.vstack((t,s)),all, title = 'Both', labeldict = labeld,size= size)
    #draw.plt.legend()

    plt.legend(markerscale=2,ncol=4,bbox_to_anchor=(1, -.12),fontsize = 'small' )
    return merged



from lmz import Map,Zip,Filter,Grouper,Range,Transpose

import cellsaw.draw
from cellsaw.similarity.measures import cosine, jaccard, precision, mkshortnames
import numpy as np
from ubergauss import tools as ut
import pandas as pd
import logging
import time

def matrixmap(method, instances, repeats = 5):
    l = len(instances)
    res = np.zeros((l,l,repeats))
    for i,obj_i in enumerate(instances):
        for j,obj_j in enumerate(instances[i:]):
            r = [method(obj_i,obj_j,x) for x in range(repeats)]
            for x,val in enumerate(r):
                res[i,j,x] = val
                res[j,i,x] = val
    return res


def matrixmap_mp(method, instances, repeats = 5):
    # pool.maxtasksperchild
    l = len(instances)
    res = np.zeros((l,l,repeats))

    def func(stuff):
        a,b,seeds,i,j = stuff
        r = [ method(a,b,seed)  for seed in seeds ]
        return r,i,j

    tmptasks =([instances[i],instances[j],list(range(repeats)), i,j]
               for i in range(l) for j in range(i+1,l))

    for r,i,j in ut.xmap(func,tmptasks):
        for x,val in enumerate(r):
            res[i,j,x] = val
            res[j,i,x] = val

    return res


def getNeighbors(matrix, labels, k = 2 ):
    sm = ut.spacemap(labels)
    codes = getnn(matrix,k)
    res = [[sm.getitem[z] for z in a] for a in codes]
    return res

def getnn(m,n):
    np.fill_diagonal(m, np.NINF)
    srt=  np.argsort(m, axis= 1)
    return [ [i]+srt[i,-n:].tolist() for i in range(srt.shape[0]) ]


def neighbors(loader, sim = cosine , filenames=[], kNeighbors = 1):
    simFunc = lambda a,b,c: sim(loader(a),loader(b))
    distances = matrixmap_mp(simFunc,filenames,repeats = 2)
    distances = np.mean(distances, axis =2)
    return getNeighbors(distances,filenames,k=kNeighbors)


def matrixmap_odd(method, targets,sources, repeats = 5):
    # instances not symmetrical
    l = len(targets)
    l2 = len(sources)
    res = np.zeros((l,l2,repeats))

    def func(stuff):
        a,b,seeds,i,j = stuff
        r = [ method(a,b,seed)  for seed in seeds ]
        return r,i,j

    tmptasks =([targets[i],sources[j],list(range(repeats)), i,j]
               for i in range(l) for j in range(l2))

    #for r,i,j in ut.xmap(func,tmptasks):
    for r,i,j in map(func,tmptasks):
        for x,val in enumerate(r):
            res[i,j,x] = val

    return res



'''
ranked_datasets_list, similarity_df = rank_by_similarity(target = target_datasets,
                                    source = source_datasets,
                                    return_similarity = True)
'''

from cellsaw.preprocess import annotate_genescore_single




from collections import defaultdict

def rank_by_sim_splitbyname(datasets, names, **kwargs):
    sn = mkshortnames(names)
    d = defaultdict(list)
    for name,item in zip(sn,datasets):
        d[name].append(item)


    if kwargs.get('return_similarity', True):
        print ('well just turn return_similarity off :)')
        kwargs['return_similarity'] = False

    flattened_results = [ line for data in d.values()
             for line in rank_by_similarity(target=data,source=data,**kwargs)]

    return flattened_results







def rank_by_similarity(target = False,
                        source = False,
                        numgenes = 2500,
                        similarity = 'jaccard',
                        dataset_name_field ='tissue5id',
                        return_similarity = True, method = 'seurat_v3'):
    '''
    target: the ones we want to annotate
    source: the database
    return_similarity: returns (list, similarity_matrix) otherwise only the list
    '''
    starttime = time.time()
    #source = ut.xmap(lambda x: annotate_genescores(x,selector=method), source)
    source = [annotate_genescore_single(s,selector = method) for s in source]
    target = [annotate_genescore_single(t,selector = method) for t in target]
    # print('got target')
    logging.info(f'obtained genescores {time.time()-starttime}')
    # source = Map(annotate_genescores, source)
    # target = Map(annotate_genescores, target)


    #breakpoint()

    if similarity  == 'cosine':
        ff = lambda a,b,c: cosine(a,b, numgenes=numgenes, scores = method)
    elif similarity == 'jaccard':
        ff = lambda a,b,c: jaccard(a,b, numgenes=numgenes, scores =method)
    else:
        raise('similarity should be either cosine or jaccard')

    distances = matrixmap_odd(ff,target,source,repeats = 1)
    distances = np.mean(distances, axis =2)
    ranksim = np.argsort(-distances)
    ranklist = [[source[col] for col in row]for row in ranksim]

    logging.info(f'did matrixmap {time.time()-starttime}')

    def getname(ada):
        return ada.uns.get(dataset_name_field,'no name')

    ind = Map(getname,target)
    col = Map(getname,source)

    distances = pd.DataFrame( distances, index = ind, columns = col)


    logging.info(f'nearly done...{time.time()-starttime}')
    if return_similarity:
        return ranklist, distances
    else:
        return ranklist






def plot_dendrogram(similarity_df):
    cellsaw.draw.dendro_degen(similarity_df.to_numpy(), similarity_df.columns, similarity_df.index)









import matplotlib.pyplot as plt
import seaborn as sns
from scipy.cluster import hierarchy as hira
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage
from sklearn.metrics import precision_score
import pandas as pd
import matplotlib.pyplot as plt
from ubergauss import tools
import numpy as np
from umap import UMAP
import seaborn as sns
from scipy.optimize import linear_sum_assignment as lsa
from lmz import *

col = plt.cm.get_cmap('tab20').colors
col = col + col + col + ((0, 0, 0),)


# col = plt.cm.get_cmap('prism')
# col = [col(i) for i in np.linspace(0, 1, 35)]
# col = col+[(0,0,0)]


def scatter(X, Y,
             title="No title",
             title_size=10,
             acc: "y:str_description" = {},
             markerscale=4,
             getmarker=lambda cla: {"marker": 'o'},
             col=col,
             label=None,
             alpha=None,
             legend=False,
             labeldict={},
             size=None):
    plt.title(title, size=title_size)
    Y = np.array(Y)
    size = max(int(4000 / len(Y)), 1) if not size else size
    plt.gca().axes.yaxis.set_ticklabels([])
    plt.gca().axes.xaxis.set_ticklabels([])
    plt.tick_params(left=False)
    plt.tick_params(bottom=False)

    if labeldict:
        for cla in np.unique(Y):
            plt.scatter(X[Y == cla, 0],
                        X[Y == cla, 1],
                        color=col[cla],
                        s=size,
                        marker=f"${cla}$",
                        edgecolors='none',
                        alpha=alpha,
                        label=labeldict.get(cla, str(cla)))  # str(cla)+" "+acc.get(cla,''),**getmarker(col[cla]))
    else:
            plt.scatter(X[:, 0], X[:, 1], c=Y, s=size)
    # plt.axis('off')
    # plt.xlabel('UMAP 2')
    # plt.ylabel('UMAP 1')
    if legend:
        # plt.legend(markerscale=2, ncol=2, bbox_to_anchor=(1, -.12))
        plt.legend(bbox_to_anchor=(1, 1), loc="upper left", markerscale=1.2, fontsize=3.5)


class tinyUmap():
    def __init__(self, dim=(3, 3), size=2, lim=False):
        figs = (size * dim[1], size * dim[0])
        plt.figure(figsize=figs, dpi=300)
        self.i = 0
        self.dim = dim
        if lim:
            concatX = np.vstack(lim)
            xmin, ymin = concatX.min(axis=0)
            xmax, ymax = concatX.max(axis=0)

            def setlim():
                plt.xlim(xmin, xmax)
                plt.ylim(ymin, ymax)
            lim = setlim()
        #self.lim = lim if lim else lambda: 0
        self.lim = setlim if lim else lambda: 0

    def next(self):
        self.i = self.i + 1
        plt.subplot(*self.dim, self.i)

    def draw(self, *a, **b):
        self.next()
        self.lim()
        scatter(*a, **b)




def plot_X(Xlist, labels, plotsperline=3,
           grad=False, size=3.5, plug = False,
           mkmix = False, mixlabels = [], titles = []):

    if not titles:
        titles = Map(str, Range(Xlist))
    # make a tinyumap with the right dimensions
    itemstodraw = len(Xlist) + mkmix
    rows = ((itemstodraw - 1) // plotsperline) + 1
    columns = plotsperline if itemstodraw > plotsperline else itemstodraw
    d = tinyUmap(dim=(rows, columns), size=size, lim = Xlist)  # default is a row

    alllabels = np.concatenate(labels)
    themap = tools.spacemap(np.unique(alllabels)) if not grad else {}

    for x, y, title in zip(Xlist, labels, titles):
        y = themap.encode(y)
        d.draw(x, y, title=title, labeldict=themap.getitem, legend = True)


    if mkmix:
        #mixlabels = [i*2 for i,stack in enumerate(Xlist) for item in stack]
        d.draw(np.vstack(Xlist),themap.encode(alllabels), labeldict= themap.getitem, legend = True)

    plt.show()




class drawMovingCenters:
    def __init__(self,XX,yy):
        # we build something like:
        # clusterid -> center1 False center3
        self.clustercenters ={}
        self.numframes = len(XX)
        self.frameid = 0

        def getacenter(x,y,label):
            y= y.values
            if label not in y:
                return False
            return x[y==label].mean(axis=0)

        for label in np.unique(np.concatenate(yy)):
            self.clustercenters[label] = [getacenter(x,y,label) for x,y in zip(XX,yy)]

    def draw(self, labelmap):
        # now we draw a frame
        alphas = 1/2**np.abs(np.array(Range(self.numframes))-self.frameid)

        for label in self.clustercenters:
            for i,(coordinate, alpha) in enumerate(zip(self.clustercenters[label][:-1],alphas)):
                if not isinstance( coordinate, bool) and not isinstance(self.clustercenters[label][i+1],bool):
                    intlabel = labelmap.getint.get(label)
                    plt.plot((coordinate[0],self.clustercenters[label][i+1][0]),
                             (coordinate[1], self.clustercenters[label][i+1][1]) , alpha = alpha, color = col[intlabel])

        for label in self.clustercenters:
            for coordinate, alpha in zip(self.clustercenters[label],alphas):
                if not isinstance( coordinate, bool):
                    intlabel = labelmap.getint.get(label)
                    plt.text(x=coordinate[0], y= coordinate[1],s= intlabel, alpha = alpha, color = col[intlabel])
        self.frameid += 1



def mkconfusion(y1, y2):
    assert len(y1) == len(y2)
    sm2 = tools.spacemap(np.unique(y2))
    sm1 = tools.spacemap(np.unique(y1))
    res = np.zeros((sm1.len, sm2.len))
    for a, b in zip(y1, y2):
        res[sm1.getint[a], sm2.getint[b]] += 1
    return res, sm1, sm2


def confusion_matrix(y1, y2, norm=True, draw=False):
    '''
    the task here is to draw a confision matrix
    returns translated labels for y2
    '''
    cm, sm1, sm2 = mkconfusion(y1, y2)


    itemCount2 = dict(zip(*np.unique(y2, return_counts=True)))
    itemCount1 = dict(zip(*np.unique(y1, return_counts=True)))

    if norm:
        cm = cm.astype(float)
        for x in Range(itemCount2):
            for y in Range(itemCount1):
                res = (2 * cm[y, x]) / (itemCount2[sm2.getitem[x]] + itemCount1[sm1.getitem[y]])
                # print(cm[y,x])
                # print(xcounts[x],ycounts[y])
                cm[y, x] = res


    names1, new_order2 = lsa(cm, maximize=True)
    # print("hung matches:",names1,new_order2, itemCount2.keys())

    if draw:
        # new_order = np.argsort(new_order2)
        # cm[names1] = cm[new_order2]

        ylab = np.unique(y1)
        for n1, n2 in zip(names1, new_order2):
            cm[[n1, n2]] = cm[[n2, n1]]
            ylab[[n1, n2]] = ylab[[n2, n1]]

        sns.heatmap(cm, xticklabels=np.unique(y2), yticklabels=ylab, annot=False, linewidths=.5, cmap="YlGnBu",
                    square=True)
        plt.xlabel('Clusters data set 2')
        plt.ylabel('Clusters data set 1')

    g = dict(zip(new_order2, names1))
    return [sm1.getitem[g[sm2.getint[y]]] for y in y2]


def plot_confusion_matrix_twice(labels):
    '''
    does the confusionmatrix twice, normalized and unnormalized, and plots!
    '''
    y1, y2 = labels
    f = plt.figure(figsize=(16, 8))
    ax = plt.subplot(121, title='absolute hits')
    confusion_matrix(y1, y2, norm=False, draw=True)
    ax = plt.subplot(122, title='relative hits (2x hits / sumOfLabels)')
    confusion_matrix(y1, y2, norm=True, draw=True)
    plt.tight_layout()




from cellsaw.merge.diffusion import kernel
from sklearn.manifold import MDS
from sklearn.decomposition import PCA
def setd2_kernel_mds(m):
    data  = kernel.linear_assignment_kernel(*m.d10, neighbors= 5,
                                            neighbors_inter=5,
                                            linear_assignment_factor=5,
                                            return_dm=True)
    r  = MDS(dissimilarity='precomputed').fit_transform(data)
    m.d2 = r[:1000], r[1000:]
    return




from lmz import Map,Zip,Filter,Grouper,Range,Transpose
from cellsaw.io_utils.loadadata import get41names, get100names, load100
from cellsaw.preprocess import annotate_genescore_single
from scipy.sparse import csr_matrix
import pandas as pd


def easyLoad100(name, path = None, remove_unlabeled = False, mingenes= 200,subsample = None,
                preprocessingmethod = 'natto', donormalize= True,
                plot=False, nattoargs=  {'mean': (0.015, 4), 'bins': (.25, 1)}):
    '''this would be the sane way to load data, but we dont use this anymore... '''
    adata = load100(name, path=path, remove_unlabeled=remove_unlabeled, subsample = subsample)
    gs = annotate_genescore_single(adata, mingenes=mingenes, selector=preprocessingmethod,
            donormalize= donormalize, nattoargs= nattoargs, plot=plot)
    return gs

import anndata
import glob
import cellsaw.io_utils.loadadata as ldata
from ubergauss import tools as t
import os
import scanpy as sc



def read(dir, suffix = '.gz',
        datasets=[],
         sampleseed = None,
        delimiter= '\t',
        sample_size = 0,
         min_genes = 200,
        remove_cells = {}):

    # targets = list of files to load
    if dir[-1] != '/':
        dir+='/'
    if not datasets:
        targets = glob.glob(f'{dir}*{suffix}')
    else:
        targets  = [f'{dir}{e}{suffix}' for e in datasets]


    # if the files are csv: convert
    if not suffix.endswith('h5'):
        t.xmap(lambda x: readcsv(x, delimiter),targets,n_jobs = 10)
        return read(dir,'.h5')


    # read an h5 file
    def openh5(fname):
        try:
            adata = anndata.read_h5ad(fname, backed = None)
            sc.pp.filter_cells(adata, min_genes=min_genes, inplace=True)
        except:
            print('basic h5 loading failed:', fname)
            return 0

        # reduce cell count
        for k,v in remove_cells.items():
            try:
                for delete_label in v:
                    adata = adata[adata.obs[k]!=delete_label]
            except:
                print('deletelabel failed in:', fname)

        if sample_size:
            try:
                sc.pp.subsample(adata, fraction=None, n_obs=sample_size,
                        random_state=sampleseed,copy=False)
            except Exception as e:
                print(e)
                print  (f"COULD NOT SUBSAMPLE {sample_size} items\
                        from {adata.uns['fname']} cells(labeled)= {adata.X.shape}")
                return adata

        # done
        print(".", end = '')
        return adata

    return Map(openh5, targets)






def readcsv(x,delimiter = '\t', saveas = ''):
    #adata = anndata.read_csv(x,delimiter=delimiter)
    things = pd.read_csv(x, sep='\t').T
    adata = anndata.AnnData(things)
    adata.X=csr_matrix(adata.X)
    path_nosuffix = str(x[:x.find('.')])
    print (path_nosuffix)

    try:
        tis__cellcnt = path_nosuffix[path_nosuffix.rindex('/')+1:].split('_')
        adata.uns['tissue'] = tis__cellcnt[0]
        adata.uns['tissue5'] = tis__cellcnt[0][:5]
        adata.uns['tissue5id'] = f'{tis__cellcnt[0][:5]}_{tis__cellcnt[-1]}'
    except:
        print("something went wrong finding tissue, tissue5 and/or tissue5id to write to 'uns'")


    adata.uns['fname'] = path_nosuffix
    if saveas:
        adata.write(saveas, compression='gzip')
    else:
        adata.write(str(path_nosuffix)+'.h5', compression='gzip')
    adata.file.close()
    #os.remove(filename)
    return 0





def annotate(adatas,annotate_function, **kwargs):
    f = annotate_function(**kwargs)
    return Map(f, adatas)


def annotate_from_barcode_csv(input_field = False, output_field = 'clusterid'):

    def annotate(adata):
        # add cluster id
        #fname = f"{path}/{adata.filename}.cluster.txt"
        fname = f"{adata.uns['fname']}.cluster.txt"
        lol = open(fname,'r').readlines()
        barcode_cid={}
        for line in lol:
            bc,cl =  line.strip().split()
            barcode_cid[bc]= int(cl)
        if not input_field or input_field == 'barcode':
            cells = adata.obs.index
        else:
            cells = adata.obs[input_field]
        adata.obs[output_field] = [barcode_cid.get(a,-1)  for a in cells]
        return adata

    return annotate

def annotate_celltypes( path = '', input_field = '', output_field = ''):
    plabels = ldata.loadpangalolabels(path)

    return lambda x: ldata.annotatetruecelltype(plabels, x, x.uns['fname'],
            f_out = output_field, f_in = input_field)


def save(adatas, format = 'h5'):
    assert format == 'h5', 'there is no generic writer in the libs currently'
    for adata in adatas:
        adata.write(adata.uns['fname']+'.h5', compression='gzip')




def nuread(dir,
        dataset,
        sampleseed = None,
        delimiter= '\t',
        sample_size = 0,
        min_genes = 200,
        remove_cells = {}):
    '''
    the other read version is annoying, we need a version that is just better :D..

    1. if there is no .h5:
        - load, annotate, save

    2. load .h5, filter, return
    '''
    expect = os.path.join(dir,'xmas',dataset+'.h5')
    if not os.path.exists(expect):
        csvpath = os.path.join(dir,dataset+'.counts.gz')
        readcsv(csvpath, delimiter,saveas = expect)
        adata = anndata.read_h5ad(expect, backed = None)

        adata = annotate_from_barcode_csv(input_field='barcode',output_field='clusterid')(adata)
        adata = annotate_celltypes(path = dir,input_field='clusterid',output_field='celltype')(adata)

        adata.write(expect, compression='gzip')

        adata.file.close()

    try:
        adata = anndata.read_h5ad(expect, backed = None)
        sc.pp.filter_cells(adata, min_genes=min_genes, inplace=True)
    except:
        print('basic h5 loading failed:', expect)
        return 0

    # reduce cell count
    for k,v in remove_cells.items():
        try:
            for delete_label in v:
                adata = adata[adata.obs[k]!=delete_label]
        except Exception as e:
            print('deletelabel failed in:', expect, e )

    if sample_size:
        try:
            sc.pp.subsample(adata, fraction=None, n_obs=sample_size,
                    random_state=sampleseed,copy=False)
        except Exception as e:
            print(e)
            print  (f"COULD NOT SUBSAMPLE {sample_size} items\
                    from {adata.uns['fname']} cells(labeled)= {adata.X.shape}")
            return adata

    if 'celltype' not in adata.obs:
        print(f'we should delete {expect}')
    return adata






import matplotlib.pyplot as plt
import seaborn as sns
from scipy.cluster import hierarchy as hira
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage
from sklearn.metrics import precision_score
import pandas as pd
import matplotlib.pyplot as plt
from ubergauss import tools
import numpy as np
from umap import UMAP
import seaborn as sns
from scipy.optimize import linear_sum_assignment as lsa
from lmz import *

col = plt.cm.get_cmap('tab20').colors
col = col + col + col + ((0, 0, 0),)


# col = plt.cm.get_cmap('prism')
# col = [col(i) for i in np.linspace(0, 1, 35)]
# col = col+[(0,0,0)]


def plot_merge(merge, labels, plotsperline=3, grad=False, size=3.5, plug = False, mkmix = False, mixlabels = []):
    '''scatterplots for merge.d2'''

    # make a tinyumap with the right dimensions

    X = merge.d2

    itemstodraw = len(X) + mkmix
    rows = ((itemstodraw - 1) // plotsperline) + 1
    columns = plotsperline if itemstodraw > plotsperline else itemstodraw

    d = tinyUmap(dim=(rows, columns), size=size)  # default is a row

    # set same limit for all the plots
    concatX = np.vstack(X)
    xmin, ymin = concatX.min(axis=0)
    xmax, ymax = concatX.max(axis=0)

    # xdiff = np.abs(xmax - xmin)
    # ydiff = np.abs(ymax - ymin)
    # plt.xlim(xmin - 0.1 * xdiff, xmax + 0.1 * xdiff)
    # plt.ylim(ymin - 0.1 * ydiff, ymax + 0.1 * ydiff)
    themap = tools.spacemap(np.unique(np.concatenate(labels)))
    for x, y in zip(X, labels):
        #y,sm = tools.labelsToIntList(y)
        y = themap.encode(y)
        if not grad:
            d.draw(x, y, title=None, labeldict=themap.getitem)
            # plt.legend(markerscale=1.5, fontsize=4, ncol=int(len(X) * 2.5), bbox_to_anchor=(1.1, -.01))
            # plt.legend(markerscale=1.5, fontsize=4, ncol=int(len(X) * 2.5))
            plt.legend(bbox_to_anchor=(1, 1), loc="upper left", markerscale=1.2, fontsize=3.5)
        else:
            d.next()
            plt.gca().axes.yaxis.set_ticklabels([])
            plt.gca().axes.xaxis.set_ticklabels([])
            plt.scatter(x[:, 0], x[:, 1], c=y, s=1)
            # plt.colorbar(shrink=.5)
            # plt.tick_params(labelsize=4)
        if plug:
            plug.draw(themap)
        plt.xlim(xmin, xmax)
        plt.ylim(ymin, ymax)


    if mkmix:
        if not mixlabels:
            mixlabelslabels = [i*2 for i,stack in enumerate(X) for item in stack]
        else:
            mixlabels = themap.encode(mixlabels)
        d.draw(np.vstack(merge.d2),mixlabels)
        plt.legend(bbox_to_anchor=(1, 1), loc="upper left", markerscale=1.2, fontsize=3.5)

    plt.show()


def tinyumap(X, Y,
             title="No title",
             title_size=10,
             acc: "y:str_description" = {},
             markerscale=4,
             getmarker=lambda cla: {"marker": 'o'},
             col=col,
             label=None,
             alpha=None,
             legend=False,
             labeldict={},
             size=None):
    plt.title(title, size=title_size)
    Y = np.array(Y)
    size = max(int(4000 / Y.shape[0]), 1) if not size else size
    embed = X
    plt.gca().axes.yaxis.set_ticklabels([])
    plt.gca().axes.xaxis.set_ticklabels([])
    plt.tick_params(left=False)
    plt.tick_params(bottom=False)
    for cla in np.unique(Y):
        plt.scatter(embed[Y == cla, 0],
                    embed[Y == cla, 1],
                    color=col[cla],
                    s=size,
                    marker=f"${cla}$",
                    edgecolors='none',
                    alpha=alpha,
                    # label= labeldict.get(cla,str(cla)), **getmarker(cla)) #str(cla)+" "+acc.get(cla,''),**getmarker(col[cla]))
                    label=labeldict.get(cla, str(cla)))  # str(cla)+" "+acc.get(cla,''),**getmarker(col[cla]))
    # plt.axis('off')
    # plt.xlabel('UMAP 2')
    # plt.ylabel('UMAP 1')
    if legend:
        plt.legend(markerscale=2, ncol=2, bbox_to_anchor=(1, -.12))


class tinyUmap():

    def __init__(self, dim=(3, 3), size=2, lim=False):
        figs = (size * dim[1], size * dim[0])
        plt.figure(figsize=figs, dpi=300)
        self.i = 0
        self.dim = dim
        if lim:
            concatX = np.vstack(lim)
            xmin, ymin = concatX.min(axis=0)
            xmax, ymax = concatX.max(axis=0)

            def setlim():
                plt.xlim(xmin, xmax)
                plt.ylim(ymin, ymax)

            lim = setlim()
        self.lim = lim if lim else lambda: 0

    def next(self):
        self.i = self.i + 1
        plt.subplot(*self.dim, self.i)

    def draw(self, *a, **b):
        self.next()
        self.lim()
        tinyumap(*a, **b)


class drawMovingCenters:
    def __init__(self,XX,yy):
        # we build something like:
        # clusterid -> center1 False center3
        self.clustercenters ={}
        self.numframes = len(XX)
        self.frameid = 0

        def getacenter(x,y,label):
            y= y.values
            if label not in y:
                return False
            return x[y==label].mean(axis=0)

        for label in np.unique(np.concatenate(yy)):
            self.clustercenters[label] = [getacenter(x,y,label) for x,y in zip(XX,yy)]

    def draw(self, labelmap):
        # now we draw a frame
        alphas = 1/2**np.abs(np.array(Range(self.numframes))-self.frameid)

        for label in self.clustercenters:
            for i,(coordinate, alpha) in enumerate(zip(self.clustercenters[label][:-1],alphas)):
                if not isinstance( coordinate, bool) and not isinstance(self.clustercenters[label][i+1],bool):
                    intlabel = labelmap.getint.get(label)
                    plt.plot((coordinate[0],self.clustercenters[label][i+1][0]),
                             (coordinate[1], self.clustercenters[label][i+1][1]) , alpha = alpha, color = col[intlabel])

        for label in self.clustercenters:
            for coordinate, alpha in zip(self.clustercenters[label],alphas):
                if not isinstance( coordinate, bool):
                    intlabel = labelmap.getint.get(label)
                    plt.text(x=coordinate[0], y= coordinate[1],s= intlabel, alpha = alpha, color = col[intlabel])
        self.frameid += 1



def mkconfusion(y1, y2):
    assert len(y1) == len(y2)
    sm2 = tools.spacemap(np.unique(y2))
    sm1 = tools.spacemap(np.unique(y1))
    res = np.zeros((sm1.len, sm2.len))
    for a, b in zip(y1, y2):
        res[sm1.getint[a], sm2.getint[b]] += 1
    return res, sm1, sm2


def confuse(y1, y2, norm=True, draw=False):
    '''
    the task here is to draw a confision matrix
    returns translated labels for y2
    '''
    cm, sm1, sm2 = mkconfusion(y1, y2)

    # print(cm.shape, len(np.unique(y1)),  len(np.unique(y2)))
    # print(np.unique(y1),  np.unique(y2))
    # sns.heatmap(cm,xticklabels=np.unique(y2),yticklabels=np.unique(y1), annot=True, linewidths=.5,cmap="YlGnBu" , square=True)
    # plt.show()
    # plt.close()

    itemCount2 = dict(zip(*np.unique(y2, return_counts=True)))
    itemCount1 = dict(zip(*np.unique(y1, return_counts=True)))

    if norm:
        cm = cm.astype(float)
        for x in Range(itemCount2):
            for y in Range(itemCount1):
                res = (2 * cm[y, x]) / (itemCount2[sm2.getitem[x]] + itemCount1[sm1.getitem[y]])
                # print(cm[y,x])
                # print(xcounts[x],ycounts[y])
                cm[y, x] = res

    # sns.heatmap(cm,xticklabels=np.unique(y2),yticklabels=np.unique(y1), annot=False, linewidths=.5,cmap="YlGnBu" , square=True)
    # plt.title('raw data')
    # plt.show()
    # plt.close()

    names1, new_order2 = lsa(cm, maximize=True)
    # print("hung matches:",names1,new_order2, itemCount2.keys())

    if draw:
        # new_order = np.argsort(new_order2)
        # cm[names1] = cm[new_order2]

        ylab = np.unique(y1)
        for n1, n2 in zip(names1, new_order2):
            cm[[n1, n2]] = cm[[n2, n1]]
            ylab[[n1, n2]] = ylab[[n2, n1]]

        sns.heatmap(cm, xticklabels=np.unique(y2), yticklabels=ylab, annot=False, linewidths=.5, cmap="YlGnBu",
                    square=True)
        plt.xlabel('Clusters data set 2')
        plt.ylabel('Clusters data set 1')

    g = dict(zip(new_order2, names1))
    return [sm1.getitem[g[sm2.getint[y]]] for y in y2]


def confuse2(labels):
    '''
    does the confusionmatrix twice, normalized and unnormalized, and plots!
    '''
    y1, y2 = labels
    f = plt.figure(figsize=(16, 8))
    ax = plt.subplot(121, title='absolute hits')
    confuse(y1, y2, norm=False, draw=True)
    ax = plt.subplot(122, title='relative hits (2x hits / sumOfLabels)')
    confuse(y1, y2, norm=True, draw=True)
    plt.tight_layout()


################
# similarity related
################
def dendro(simiarity, labels, distancecut = 1):
    print ('deprecated?')
    '''
    # HEATMAP
    f=plt.figure(figsize=(16,8))
    ax=plt.subplot(121)
    ax.set_title('complete similarity matrix', fontsize=20)
    sns.heatmap(similarity, xticklabels = False,
            yticklabels = labels,  cmap="YlGnBu")
    locs, labels = plt.yticks()
    plt.setp(labels,size = 8)

    # DENDROGRAM
    ax=plt.subplot(122)
    ax.set_title('induced dendrogram (ward)', fontsize=20)
    Z = squareform(similarity)
    Z = hira.linkage(1-Z,'ward')
    hira.dendrogram(Z, labels = labels, color_threshold=distancecut, orientation='right')
    locs, labels = plt.xticks()
    plt.setp(labels, rotation=90,size = 8)
    plt.subplots_adjust(wspace = .4)
    return hira.fcluster(Z, t=7, criterion = 'maxclust')
    '''


def dendro_degen(similarity, xlabels,ylabels, distancecut = 1):
    '''
    so for the new mode the matrix is not symmetrical anymore..
    '''

    f=plt.figure(figsize=(16,8))
    ax=plt.subplot(122)

    # HEATMAP
    ax.set_title('complete similarity matrix', fontsize=20)
    sns.heatmap(similarity, xticklabels = xlabels,
            yticklabels = ylabels,  cmap="hot")
    locs, labels = plt.yticks()
    plt.setp(labels,size = 8)


    # DENDROGRAM
    ax=plt.subplot(121)

    ax.set_title('induced dendrogram (ward)', fontsize=20)
    #Z = squareform(similarity)
    Z = similarity
    Z = hira.linkage(1-Z,'ward')
    hira.dendrogram(Z, labels = labels, color_threshold=distancecut, orientation='right')
    locs, labels = plt.xticks()
    plt.setp(labels, rotation=90,size = 8)
    plt.subplots_adjust(wspace = .4)


    return hira.fcluster(Z, t=7, criterion = 'maxclust')


def plotPrecision(data, wg = True, method='cosine similarity'):
    # plor precision at k :)
    # [val,rep,p@k,n_genes]
    df = pd.DataFrame(data)
    df.columns = 'precision rep neighbors±σ genes sim preproc'.split()
    title = f'Searching for similar datasets via {method}'

    if wg:
        sns.set_theme(style='whitegrid')
    sns.lineplot(data = df,x='genes', y = 'precision',style= 'neighbors±σ',
            hue = 'neighbors±σ',palette="flare", ci=68)
    plt.title(title, y=1.06, fontsize = 16)
    plt.ylim([.75,1])
    plt.ylabel('precision of neighbors (40 datasets)')
    plt.xlabel('number of genes')


import gif
def makegif(merge, labels, plug):
	gif.options.matplotlib["dpi"] = 200
	def plot_merge(merge, labels, plotsperline=3, grad=False, size=3.5, plug = False):
		X = merge.d2
		#rows = ((len(X) - 1) // plotsperline) + 1
		#columns = plotsperline if len(X) > plotsperline else len(X)
		d = tinyUmap(dim=(1, len(merge.d2)), size=size)  # default is a row
		plt.close()
		# set same limit for all the plots
		concatX = np.vstack(X)
		xmin, ymin = concatX.min(axis=0)
		xmax, ymax = concatX.max(axis=0)

		themap = tools.spacemap(np.unique(np.concatenate(labels)))
		@gif.frame
		def myplotter(x,y):
			#y,sm = tools.labelsToIntList(y)
			y = themap.encode(y)
			if not grad:
				tinyumap(x, y, title=None, labeldict=themap.getitem)
				# plt.legend(markerscale=1.5, fontsize=4, ncol=int(len(X) * 2.5), bbox_to_anchor=(1.1, -.01))
				# plt.legend(markerscale=1.5, fontsize=4, ncol=int(len(X) * 2.5))
				plt.legend(bbox_to_anchor=(1, 1), loc="upper left", markerscale=1.2, fontsize=3.5)
			else:
				plt.gca().axes.yaxis.set_ticklabels([])
				plt.gca().axes.xaxis.set_ticklabels([])
				plt.scatter(x[:, 0], x[:, 1], c=y, s=1)
				# plt.colorbar(shrink=.5)
				# plt.tick_params(labelsize=4)
			if plug:
				plug.draw(themap)
			plt.xlim(xmin, xmax)
			plt.ylim(ymin, ymax)

		frames = []
		for x, y in zip(X, labels):
			frames.append(myplotter(x,y))
		gif.save(frames, 'meganice.gif', duration=1000)
	plot_merge(merge,labels,plug=plug)



from  sklearn.neighbors import KNeighborsClassifier
from ubergauss import tools as ut
import numpy as np


def repeat_as_column(a,n):
    return np.tile(a,(n,1)).T

def neighbor_labelagreement(X_train, y_train,n_neighbors):
    knn = KNeighborsClassifier(n_neighbors=n_neighbors+1)


    y_train, _ = ut.labelsToIntList(y_train)
    y_train= np.array(y_train)
    # print(sum(np.isnan(X_train)))
    knn.fit(X_train, y_train)

    # Get the labels of the nearest neighbors for each training point
    _, indices = knn.kneighbors(X_train)
    neighbor_labels = y_train[indices]
    # Exclude the label of the training point itself
    neighbor_labels = neighbor_labels[:, 1:]

    # Compute the training error
    agreement = (repeat_as_column(y_train,n_neighbors) == neighbor_labels).mean()

    return agreement




from lmz import Map,Zip,Filter,Grouper,Range,Transpose
from scipy.sparse.csgraph import minimum_spanning_tree
from pprint import pprint
import time
from scipy.optimize import linear_sum_assignment
from sklearn import neighbors as nbrs, metrics
import numpy as np
from scipy import sparse
from scipy.sparse.csgraph import dijkstra
from ubergauss import tools
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.utils import check_symmetric


####################
#  LINEAR SUM ASSIGNMENT
####################

def iterated_linear_sum_assignment(distances, repeats):
    def linear_sum_assignment_iteration(distances):
        r, c = linear_sum_assignment(distances)
        d = distances[r,c].copy()
        distances [r,c] = np.inf
        return r,c,d

    cc, rr, dist  = Transpose([ linear_sum_assignment_iteration(distances) for _ in range(repeats)  ])
    return np.hstack(cc) , np.hstack(rr), np.hstack(dist)



def linear_sum_assignment_matrices(x1, x2, repeats,
                                   dist = True,
                                   dense = True):
    x= metrics.euclidean_distances(x1,x2)
    a,b, distances = iterated_linear_sum_assignment(x, repeats)
    r = np.zeros(x.shape) if dense else sparse.csr_matrix(x.shape, dtype=np.float32)
    # print(f"{a.shape=} {b.shape=} {distances.shape=}]")
    # print(f"{a=} {b=} {distances=}]")
    r[a,b] = (distances + 0.0001) if dist else 1
    return r,r.T

#######
# 1nn distance
###########
def avg1nndistance(listofdistancemat):
    def calculate1nn_distance(q):
        return np.mean([ np.min(q[q!=0])  for row in q])
    list1nn_means = Map(calculate1nn_distance,listofdistancemat)
    return np.mean(list1nn_means)


###################
# graphbla
##################


def neighborgraph(x, neighbors):
    z= nbrs.kneighbors_graph(x,neighbors,mode='distance')
    diff = z-z.T
    diff[diff > 0 ] = 0
    z-= diff
    check_symmetric(z,raise_exception=True)

    return z


from matplotlib import pyplot as plt
def linear_assignment_kernel(x1,x2, neighbors = 3,
                             neighbors_inter= 1, sigmafac = 1, linear_assignment_factor = 1, return_dm = False):


    '''
    X are the stacked projections[0] (normalized read matrices)
    since this is a kernel, we return a similarity matrix

    - we do neighbors to get quadrant 2 and 4
    - we do hungarian to do quadrants 1 and 3
    - we do dijkstra to get a complete distance matrix

    - dijkstra wants dense matrices, so we can not go sparse

    '''
    q1,q3 = linear_sum_assignment_matrices(x1,x2, neighbors_inter,
                                           dist = True,
                                           dense = True)

    q2,q4 = [neighborgraph(x,neighbors).todense() for x in [x1,x2]]


    q1 = q1*linear_assignment_factor
    q3 = q3*linear_assignment_factor

    #distance_matrix = sparse.hstack((sparse.vstack((q2,q3)),sparse.vstack((q1,q4)))).todense()
    distance_matrix = np.hstack((np.vstack((q2,q3)),np.vstack((q1,q4))))


    # print(f"raw dist")
    # sns.heatmap(distance_matrix);plt.show()


    distance_matrix = dijkstra(distance_matrix, directed = False)

    if return_dm:
        return distance_matrix

    dijkstraQ1 = distance_matrix[:x1.shape[0],x1.shape[0]:]

    sigma = avg1nndistance([q2,q4])*sigmafac
    similarity_matrix = np.exp(-dijkstraQ1/sigma)

    # print('dijkstra zoom');sns.heatmap(dijkstraQ1); plt.xlabel('target'); plt.show()
    # print(f'gaussed  sigme:{sigma}')
    # sns.heatmap(similarity_matrix); plt.show()

    return  similarity_matrix

def linear_assignment_kernel_XXX(x1,x2, neighbors = 3,
                                 neighbors_inter= 1, sigmafac = 1, linear_assignment_factor = 1, tsizes = None):


    '''
    now we have multiple target datasets:
        i guess we build the target line by line
    '''

    # rist we split x2:

    dslist = [x1]+np.split(x2,np.add.accumulate(tsizes))

    if dslist[2].shape[0] == 0: # happens on self similarity
        dslist = dslist[:2]
    if dslist[-1].shape[0] == 0: # np splitting likes to add 0-frames when things fit perfectly :(
        dslist = dslist[:-1]

    # then we built a row:
    diaglist = []
    row = []
    for i in range(len(dslist)):
        col = []
        for j in range(len(dslist)):
            if i == j:
                block = neighborgraph(dslist[i],neighbors).todense()
                diaglist.append(block)
            if (i+1) == j:
                block,_ = linear_sum_assignment_matrices(dslist[i],dslist[j], neighbors_inter, dist = True, dense = True)
                block*=linear_assignment_factor
            if (j+1) == i:
                _,block = linear_sum_assignment_matrices(dslist[i],dslist[j], neighbors_inter, dist = True, dense = True)
                block*=linear_assignment_factor
            else:
                np.zeros((dslist[i].shape[0],dslist[j].shape[0]))
            col.append(block)
        row.append(np.hstack(col))
    distance_matrix = np.vstack(row)
    distance_matrix = dijkstra(distance_matrix, directed = False)
    dijkstraQ1 = distance_matrix[:x1.shape[0],x1.shape[0]:]
    sigma = avg1nndistance(diaglist)*sigmafac
    similarity_matrix = np.exp(-dijkstraQ1/sigma)

    # print('dijkstra zoom');sns.heatmap(dijkstraQ1); plt.xlabel('target'); plt.show()
    # print(f'gaussed  sigme:{sigma}')
    # sns.heatmap(similarity_matrix); plt.show()

    return  similarity_matrix





from sklearn.neighbors import NearestNeighbors
from cellsaw.merge import mergehelpers


def avgdist(a,numneigh = 2):
    nbrs = NearestNeighbors(n_neighbors=1+numneigh).fit(a)
    distances, indices = nbrs.kneighbors(a)
    return np.mean(distances[:,1:], axis = 1)

def average_knn_distance(I,J,i_ids,j_ids,numneigh):
    d1  = avgdist(I,numneigh)[i_ids]
    d2  = avgdist(J,numneigh)[j_ids]
    stack = np.vstack((d1,d2))
    return np.mean(stack, axis=0).T

from sklearn.neighbors import NearestNeighbors
def neighborgraph_p_weird(x, neighbors):
    # neighbors = max(1, int(x.shape[0]*(neighbors_perc/100)))
    z= nbrs.kneighbors_graph(x,neighbors)
    diff = z-z.T
    diff[diff > 0 ] = 0
    z-= diff
    return z

def neighborgraph_p_real(x, neighbors):
    z = np.zeros_like(x)
    np.fill_diagonal(x,np.NINF)
    for i,row in enumerate(x):
        sr = np.argsort(row)
        z[i][sr[-neighbors:]]  = 1
    diff = z-z.T
    diff[diff > 0 ] = 0
    z-= diff
    return z





def make_adjacency(similarity, algo=0, neighbors=10):
        n_perc = neighbors = max(1, int(similarity.shape[0]*(neighbors/100)))
        simm = neighborgraph_p_weird if algo == 1 else neighborgraph_p_real
        return simm(similarity, n_perc)

def merge_adjacency(*args):
    if len(args) == 1:
        return args[0]
    return np.logical_or(args[0],merge_adjacency(*args[1:]))

def make_star(size = 5, center =2):
    ret = np.zeros((size,size))
    ret[center] = 1
    ret[:,center] = 1
    return ret

def make_sequence(size = 5, indices = [0,1]):
    ret = np.zeros((size,size))
    for i in range(size):
        for j in indices:
            if i+j < size and i+j >= 0:ret[i,i+j] = 1
    return ret


def test_matrixmaker():
    m = make_star(size = 5, center = 1)
    s = make_sequence(size=5, indices=[0,1,2])
    print(s,m,merge_adjacency(s,m))


def linear_assignment_integrate(Xlist,
                                intra_neigh=15,
                                inter_neigh = 1,
                                scaling_num_neighbors = 2,
                                outlier_threshold = .8,
                                scaling_threshold=.9,
                                dataset_adjacency = False,
                                showtime = False):

    lsatime = 0.0
    eutime = 0.0



    def adjacent(i,j):
        if isinstance( dataset_adjacency, np.ndarray):
            return  dataset_adjacency[i][j] == 1
        else:
            return True

    def make_distance_matrix(i,j):
            if i == j:
                # use the maximum between neighborgraph and min spanning tree to make sure all is connected
                neighborAdj = sparse.lil_matrix(neighborgraph(Xlist[i],intra_neigh))
                distancemat = metrics.euclidean_distances(Xlist[i])
                tree = sparse.lil_matrix(minimum_spanning_tree(distancemat))
                return tree.maximum(neighborAdj), 0,0


            elif not adjacent(i,j):
                # just fill with zeros :)
                return  sparse.lil_matrix((Xlist[i].shape[0],Xlist[j].shape[0]), dtype=np.float32),0,0
            else:
                '''
                based = hungdists/= .25
                dm = avg 1nn dist in both
                based*=dm
                '''
                # hungarian
                eustart=time.time()
                ij_euclidian_distances= metrics.euclidean_distances(Xlist[i],Xlist[j])
                eutime = time.time() - eustart
                res = sparse.lil_matrix(ij_euclidian_distances.shape,dtype=np.float32)
                if inter_neigh==0:
                    return res

                lsastart=time.time()
                i_ids,j_ids, ij_lsa_distances = iterated_linear_sum_assignment(ij_euclidian_distances,inter_neigh)
                lsatime=(time.time()-lsastart)

                # remove worst 25% hits
                sorted_ij_assignment_distances  = np.sort(ij_lsa_distances)
                lsa_outlier_thresh = sorted_ij_assignment_distances[int(len(ij_lsa_distances)*outlier_threshold)]
                outlier_ids = ij_lsa_distances >  lsa_outlier_thresh
                ij_lsa_distances[outlier_ids] = 0

                # normalize
                lsa_normalisation_factor = sorted_ij_assignment_distances[int(len(ij_lsa_distances)*scaling_threshold)]
                ij_lsa_distances /= lsa_normalisation_factor
                #
                #dm = (avgdist(Xlist[i], hoodsize)+avgdist(Xlist[j],hoodsize))/2
                #dab *=dm
                average_knn_distance_factors = average_knn_distance(Xlist[i],Xlist[j], i_ids , j_ids, scaling_num_neighbors)
                ij_lsa_distances *= average_knn_distance_factors


                # make a matrix
                res[i_ids,j_ids] = ij_lsa_distances
                return res, lsatime, eutime


    # then we built a row:
    row = []
    for i in range(len(Xlist)):
        col = []
        for j in range(len(Xlist)):
            if i <= j:
                distance_matrix,a,b = make_distance_matrix(i,j)
                lsatime+=a
                eutime += b
            else:
                distance_matrix = row[j][i].T

            col.append(distance_matrix)
        row.append(col)


    distance_matrix = sparse.vstack([sparse.hstack(col) for col in row])

    if showtime:
        print(f"{eutime=}")
        print(f"{lsatime=}")
    return  distance_matrix




def KNNFormater(Data, precomputedKNNIndices, precomputedKNNDistances):
    from pynndescent import NNDescent
    # print('asd1')
    pyNNDobject = NNDescent(np.vstack(Data), metric='euclidean', random_state=1337,n_jobs = 1)
    # print('asd2')
    pyNNDobject._neighbor_graph = (precomputedKNNIndices.copy(), precomputedKNNDistances.copy())
    precomputedKNN = (precomputedKNNIndices, precomputedKNNDistances, pyNNDobject)
    return precomputedKNN

def stack_n_fill(a,val):
    '''
    reformats the data
    '''
    maxlen = max(Map(len,a))
    res = np.full((len(a),maxlen),val)
    for i,val in enumerate(a):
        res[i,:len(val)] = val
    return res



from umap import UMAP
def distmatrixumap(dataXlist,dm,components = 10):
    sparseMatrix = sparse.csr_matrix(dm)
    precomputedKNNIndices = []
    precomputedKNNDistances = []
    # for ip in range(len(sparseMatrix.indptr)-1):
    #         start = sparseMatrix.indptr[ip]
    #         end = sparseMatrix.indptr[ip+1]
    #         precomputedKNNIndices.append(sparseMatrix.indices[start:end])
    #         precomputedKNNDistances.append(sparseMatrix.data[start:end])
    for row in sparseMatrix:
        precomputedKNNIndices.append(row.indices)
        precomputedKNNDistances.append(row.data)
    umapknn = stack_n_fill(precomputedKNNIndices,-1), stack_n_fill( precomputedKNNDistances,np.inf)


    precomputedKNN = KNNFormater(dataXlist, *umapknn)
    n_neighbors = precomputedKNN[0].shape[1]

    mymap = UMAP(n_components=components, #Dimensions to reduce to
                 n_neighbors=n_neighbors,
                 random_state=1337,
                 metric='euclidean',
                 precomputed_knn=precomputedKNN,
                 force_approximation_algorithm=True)
    r=  mymap.fit_transform(np.vstack(dataXlist))
    return r






from lmz import *
from sklearn.mixture import _gaussian_mixture as _gm
import numpy as np
from sklearn.metrics.pairwise import euclidean_distances as ed
from sklearn.mixture import GaussianMixture as gmm
from ubergauss import tools


###################
#  OPTIMIZE GMM FOR 2 DATA SETS
##################

def hot1(y):
    # be sure to start with label 0 ...
    uy =  y.max()+1
    r= np.zeros((len(y), uy))
    r[range(len(y)), y] =1
    return r


def get_means_resp(X,log_resp, cov):
    _, means_, covariances_ = _gm._estimate_gaussian_parameters(X, np.exp(log_resp), 1e-6, cov)
    precisions_cholesky_    = _gm._compute_precision_cholesky( covariances_, cov)
    log_resp                = _gm._estimate_log_gaussian_prob( X, means_, precisions_cholesky_,cov)
    return means_, log_resp



###################
#  OPTIMIZE GMM FOR MANY DATA SETS
##################


def optimize_MANY(XXX,y, cov='tied'):
    log_resp = hot1(y)  # init
    mmm, lll =  Transpose([get_means_resp(x,log_resp,cov) for x in XXX])
    #log_resp = l1+l2
    # l contains for each cluster( horizontal ) a probability for each cell( vertical )
    log_resp = np.array(lll).sum(axis=0)

    all_the_labels = [ l.argmax(axis=1)  for l in lll]
    all_the_probas = [ l.max(axis=1)  for l in lll]
    erer = [ np.logical_not(np.all( np.array(a) == a[0]  ))   for a in zip(*all_the_labels)] # are all entries the same?
    return log_resp.argmax(axis=1),erer, all_the_labels, lll

def multitunnelclust(XXX,y, method = 'full', n_iter=100, debug = False):
    '''
    method tied doesnt work?
    maybe because the covariancematrix does not have the expected shape
    '''
    for asd in range(n_iter):
        yold = y.copy()
        y , e, all_labels, all_probas =  optimize_MANY(XXX,y,cov=method)
        chang = sum(y!=yold)
        if debug > 1:
            print(f"changes in iter:  {chang}")
        if chang == 0:
            if debug:print(f"model converged after {asd} steps")
            break
    else:
        assert False, "did not converge"
    return y,e, all_labels, all_probas



def initialize(X,clusts=10, cov = 'tied'):
    return gmm(n_components=clusts, n_init=30, covariance_type=cov).fit_predict(X)



def HEM(merge, init = 0, pid =2):
    assert merge.sorted
    if isinstance(init, int):
        print("init was an int so we calculate the init vector")
        init = initialize(merge.projections[pid][0])
    return multitunnelclust(merge.projections[pid],init,method = 'full')



def linasEM(nparrays, y):
    '''
        i assume a set of string labels are given
    '''
    sm  = tools.spacemap(np.unique(y))
    intlabels = np.array(sm.encode(y))
    predicted_intlabels,_,_,_ = multitunnelclust(nparrays,intlabels)
    return sm.decode(predicted_intlabels)

























#############
# KMEANS
############
def assign(x1,x2,c1,c2):
    r = ed(x1,c1)
    r2 = ed(x2,c2)
    r3=np.concatenate((r,r2), axis=1) #euclidean distances of points
    z = np.argmin(r3, axis = 1) # cluster with minimal euclidean distance
    res = np.array( [zz if zz < c1.shape[0] else zz-c1.shape[0]  for zz in z] ) # y

    return res,  np.argmin(r, axis = 1) !=  np.argmin(r2, axis = 1)


def centers(y,X):
    cents = []
    for i in np.unique(y):
        cents.append(X[y==i].mean(axis=0))
    return np.array(cents)


def optimize_kmeans(X1,X2,y):
    c1, c2 = centers(y,X1), centers(y,X2)
    #c2 = h.hungsort(c1,c2)
    y,e = assign(X1,X2,c1,c2)
    return y,e


#######
# Multi input KMEANS
#######
def multi_assign(X, c):
    listOfRs = []
    r_n = ed(X[0], c[0])

    listOfEDs = [np.argmin(r_n, axis=1)]
    for i in range(1,len(X)):
        r_i = ed(X[i],c[i])
        r_n = np.concatenate((r_n, r_i), axis=1)
        listOfEDs.append(np.argmin(r_i, axis=1)) # Used to find error

    clustersByX = np.stack(listOfEDs, axis=0)

    z = np.argmin(r_n, axis=1)
    res = np.array( [zz if zz < c[0].shape[0] else zz-(zz//c[0].shape[0])*c[0].shape[0]  for zz in z] )

    e = np.all(clustersByX == clustersByX[0,:], axis=0)
#    e = np.argmin(ed(X[0],c[0]), axis = 1) !=  np.argmin(ed(X[1], c[1]), axis = 1)

    return res, e


def optimize_multi_kmeans(XList, y):
    cList = []
    for x in XList:
        cList.append(centers(y, x))
    y,e = multi_assign(XList, cList)
    return y,e






import numpy as np
import logging
from scipy.optimize import linear_sum_assignment
from sklearn.metrics import pairwise_distances


class Merge(mergeutils):
    def __init__(self, adatas, selectgenes = 2000,
                 make_even = True, pca = 40, umaps = [2],
                 joint_space = True,
                 sortfield = 0,
                 oldcut = False,
                 genescoresid = '',
                 titles = "ABCDEFGHIJKIJ"):

        assert isinstance(adatas, list), f'merge wants a list, not {type(adatas)}'
        shapesbevorepp= [a.X.shape for a in adatas]
        assert all([a.X.shape[1] == adatas[0].X.shape[1] for a in adatas])
        #self.genescores = [a.varm['scores'] for a in adatas]

        scorename = genescoresid or adatas[0].uns['lastscores']
        self.genescores = [a.varm[scorename] for a in adatas]
        #self.geneab = [a.varm['genes'] for a in adatas]
        if oldcut:
            self.data  = unioncut(self.genescores, selectgenes, adatas)
        else:
            self.data  = equal_contrib_gene_cut(self.genescores, selectgenes, adatas)
        self.sorted = False
        self.jointSpace = joint_space

        # geneab = np.all(np.array(self.geneab), axis=0)
        # for i, d in enumerate(self.data):
        #     data[i] = d[:, geneab]

        if make_even:
            self.data = make_even(self.data)


        logging.info("preprocess:")
        logging.info(f"{make_even=}")
        for a,b in zip(shapesbevorepp, self.data):
            logging.info(f"{a} -> {b.shape}")

        ######################
        self.titles = titles

        # do dimred
        self.projections = [[ d.X for d in self.data]]+dimension_reduction(self.data,
                                                False, # scale (will still scale if pca)
                                                False, # will be ignored anyway
                                                PCA=pca,
                                                umaps=umaps,
                                                joint_space=joint_space)

        if pca:
            self.PCA = self.projections[1]

        if sortfield >=0:
            assert make_even==True, 'sorting uneven datasets will result in an even dataset :)'
            self.sort_cells(sortfield)

        if umaps:
            for x,d in zip(umaps,self.projections[int(pca>0)+1:]):
                self.__dict__[f"d{x}"] = d


    def sort_cells(self,projection_id = 0):
        # loop over data sets
        self.hungdist =[]
        for i in range(len(self.data)-1):
            hung, dist = self.hungarian(projection_id,i,i+1)
            self.hungdist.append(dist)
            #self.data[i+1] = self.data[i+1][hung[1]]
            #self.data[i+1].X = self.data[i+1].X[hung[1]]
            for x in range(len(self.projections)):
                self.projections[x][i+1] = self.projections[x][i+1][hung[1]]
                if x == 0:
                    self.data[i+1]= self.data[i+1][hung[1]]

        self.sorted = True

    #
    def hungarian(self,data_fld,data_id, data_id2):
        hung, dist = hungarian(self.projections[data_fld][data_id],
                                            self.projections[data_fld][data_id2])
        return hung, dist[hung]




def hungarian(X1, X2, debug = False):
    distances = pairwise_distances(X1,X2, metric=metric)
    row_ind, col_ind = linear_sum_assignment(distances)
    if debug:
        x = distances[row_ind, col_ind]
        num_bins = 100
        print("hungarian: debug hist")
        plt.hist(x, num_bins, facecolor='blue', alpha=0.5)
        plt.show()
    return (row_ind, col_ind), distances


def unioncut(scores, numGenes, data):
    indices = np.argpartition(scores, -numGenes)[:,-numGenes:]
    indices = np.unique(indices.flatten())
    [d._inplace_subset_var(indices) for d in data]
    return data

def equal_contrib_gene_cut(scores,numgenes,data):
    ar = np.array(scores)
    ind = np.argsort(ar)
    def calcoverlap(scores,number):
        indices = scores[:,-number:]
        indices = np.unique(indices.flatten())
        return indices
    def findcutoff(low,high, lp = -1):
        probe = int((low+high)/2)
        y = calcoverlap(ind,probe)
        if probe == lp:
            return y
        if len(y) > numgenes:
            return findcutoff(low,probe,probe)
        else:
            return findcutoff(probe,high,probe)
    indices = findcutoff(0,numgenes)
    [d._inplace_subset_var(indices) for d in data]
    return data


def umapify(dx, dimensions,n_neighbors=10):
    mymap = umap.UMAP(n_components=dimensions,
                      n_neighbors=n_neighbors,
                      random_state=1337).fit(np.vstack(dx))
    return [mymap.transform(a) for a in dx]



def dimension_reduction(adatas, scale, zero_center, PCA, umaps):
    # get a (scaled) dx
    if scale or PCA:
        adatas= [sc.pp.scale(adata, zero_center=False, copy=True,max_value=10) for adata in adatas]
    dx = [adata.to_df().to_numpy() for adata in adatas]

    res = []
    if PCA:
        pca = decomposition.PCA(n_components=PCA)
        pca.fit(np.vstack(dx))
        #print('printing explained_variance\n',list(pca.explained_variance_ratio_))# rm this:)
        dx = [ pca.transform(e) for e in dx  ]
        res.append(dx)

    for dim in umaps:
        assert 0 < dim < PCA, f'{dim=} {PCA=}'
        res.append(umapify(dx,dim))

    return res

    def make_even(data):
        # assert all equal
        size = data[0].X.shape[1]
        assert all([size == other.X.shape[1] for other in data])

        # find smallest
        counts = [e.X.shape[0] for e in data]
        smallest = min(counts)

        for a in data:
            if a.X.shape[0] > smallest:
                sc.pp.subsample(a,
                                fraction=None,
                                n_obs=smallest,
                                random_state=0,
                                copy=False)
        return data




import numpy as np
from ubergauss import tools
from sklearn.semi_supervised import LabelSpreading,LabelPropagation
from cellsaw.merge.diffusion.kernel import linear_assignment_kernel, linear_assignment_kernel_XXX

def stringdiffuse(mergething, labels, pid = 1,
                                sigmafac = 1,
                              neighbors_inter = 1,
                              neighbors_intra = 7,linear_assignment_factor=1):

    '''
    will use the first dataset to train and return prediction on the second
    '''

    # the data i use has theese properties:
    # -1 -> no true label
    # "Unknown" -> it was decided that we treat this as a normal label
    # "pangalo error" -> does not happen often, will be treated as normal label
    # thus encoding labels like this is fine:
    sm = tools.spacemap(
            np.unique(
                [xx for xx in labels if isinstance(xx, str)]
            ))

    diffusor = Diffusion(n_neighbors_inter = neighbors_inter,
                         sigmafac= sigmafac,
                         n_neighbors_intra  = neighbors_intra,
                         linear_assignment_factor=linear_assignment_factor)
    diffusor.fit(mergething.projections[pid][1], sm.encode(labels))
    intresults =  diffusor.predict(mergething.projections[pid][0])
    return sm.decode(intresults)

class Diffusion:

    def __init__(self,
                n_neighbors_intra = 7,
                n_neighbors_inter=1,
                sigmafac = 1,
                lp_model= LabelPropagation(kernel = None, max_iter=1000),
                linear_assignment_factor = 1,
                kernel = linear_assignment_kernel):

        """we just run diffusion as sklearn would i.e. expect no string labels etc"""

        self.neighbors_intra = n_neighbors_intra
        self.neighbors_inter = n_neighbors_inter
        self.lp_model = lp_model
        self.kernel = kernel
        self.sigmafac = sigmafac
        self.linear_assignment_factor =linear_assignment_factor


    def fit(self,X,y):
        self.X = X
        self.y = np.array(y)
        self.train_ncells = X.shape[0]
        kernel = lambda x1, x2: self.kernel(x1,x2,
                                            neighbors = self.neighbors_intra,
                                            neighbors_inter = self.neighbors_inter,
                                            sigmafac = self.sigmafac,
                                            linear_assignment_factor=self.linear_assignment_factor)
        self.lp_model.set_params(kernel = kernel)


    def predict(self,X,y = False):
        if not y:
            y = np.full(X.shape[0],-1)


        #assert self.X.shape == X.shape, 'not sure why i assert this... probably because i havent tested it'
        #assert self.y.shape == y.shape, 'assert everything!'


        # print(f"FITT")
        self.lp_model.fit(self.X, self.y)
        # print(f"{self.lp_model.label_distributions_=}")
        # print(f"PREDICT")
        res = self.lp_model.predict(X)
        return res

        # Ystack = np.hstack((self.y,y))
        # Xstack = np.vstack(Map(tools.zehidense, (self.X,X)))
        # all_labels = self.lp_model.fit(Xstack, Ystack).transduction_
        # self.correctedlabels = all_labels[self.train_ncells]
        # return  all_labels[self.train_ncells:]




from lmz import Map,Zip,Filter,Grouper,Range,Transpose
from scipy.sparse import issparse
from sklearn import decomposition
import numpy as np
from scipy.optimize import linear_sum_assignment
import ubergauss.tools as ut
from lucy import draw
from sklearn.metrics import pairwise_distances
from anndata._core.merge import concat
import scanpy as sc
import umap as uumap


# plotting
def plot_confusion_matrix_normalized_raw(adatas, label = 'label', alignmentbase = 'pca40'):
    assert len(adatas) == 2
    adatas = align(adatas, base= alignmentbase)
    draw.plot_confusion_matrix_twice(*[x.obs['label'] for x in adatas])

def plot(adatas, projection = 'umap2', label= 'label', **kwargs):
    X = to_arrays(adatas, base=projection)
    labels = [a.obs[label] for a in adatas]
    batch_labels = [a.obs['batch'][0] for a in adatas]
    draw.plot_X(X, labels,titles = batch_labels,**kwargs)



# similarity
def jaccard_distance(a,b, num_genes):
    binarized_hvg = np.array([ ut.binarize(d,num_genes) for d in [a,b] ])
    union = np.sum(np.any(binarized_hvg, axis=0))
    intersect = np.sum(np.sum(binarized_hvg, axis=0) ==2)
    return intersect/union


def similarity(adatas, hvg_name = 'cell_ranger', num_genes = 2000):
    assert hvg_name in adatas[0].var, 'not annotated..'
    #genescores = [a.var[hvg_name] for a in adatas]
    genescores = [a.uns[hvg_name] for a in adatas]
    res = [[ jaccard_distance(a,b, num_genes = num_genes) for a in genescores] for b in genescores]
    return np.array(res)



# checking data
def check_adatas(adatas):
        assert isinstance(adatas, list), f'merge wants a list, not {type(adatas)}'
        assert all([a.X.shape[1] == adatas[0].X.shape[1] for a in adatas])



# sometimes we need to stack the adatas to work on the concatenated values...


# def to_arraylist(adatas):
#    # make a list of tupples for use in sklearn
#     return X,cell_name, genename

def stack(adatas):
    assert 'batch' in adatas[0].obs
    return concat(adatas)

def stack_single_attribute(adatas, attr = ""):
    if not attr:
        data = [a.X for a in adatas]
    else:
        data = [a.obsm[attr] for a in adatas]
    return ut.vstack(data)

def split_by_adatas(adatas, stack):
    batch_ids = np.hstack([ a.obs['batch'] for a in adatas])
    return [ stack [batch_ids == batch] for batch in np.unique(batch_ids)]

def attach_stack(adatas, stack, label):
    '''
    if we generate data for all cells on the stacked-adatas,
    this function can split the data and assign it to the adatas
    '''
    stack_split = split_by_adatas(adatas,stack)
    for a,s in zip(adatas,stack_split):
        a.obsm[label] = s
    return adatas



def preprocess(adatas,cut_ngenes = 2000, cut_old = False, hvg = 'cell_ranger', make_even = True):
    check_adatas(adatas)
    if hvg == 'cell_ranger':
        adatas = cell_ranger(adatas)
    else:
        assert False

    selector = hvg_ids_from_union if cut_old else hvg_ids_from_union_limit

    for a in adatas: # saving this info to be able to calculate similarity
        a.uns[hvg] = a.var[hvg]

    adatas = hvg_cut(adatas, selector(adatas,cut_ngenes,hvg_name=hvg))
    if make_even:
        adatas = subsample_to_min_cellcount(adatas)
    return adatas

def hvg_cut(adatas,hvg_ids):
    [d._inplace_subset_var(hvg_ids) for d in adatas]
    return adatas

def hvg_ids_from_union(adatas, numGenes, hvg_name= 'cell_ranger'):
    scores = [a.var[hvg_name] for a in adatas]
    hvg_ids_per_adata = np.argpartition(scores, -numGenes)[:,-numGenes:]
    hvg_ids = np.unique(hvg_ids_per_adata.flatten())
    return hvg_ids


def hvg_ids_from_union_limit(adatas,numgenes,hvg_name = 'cell_ranger'):

    scores = [a.var[hvg_name] for a in adatas]
    ar = np.array(scores)
    ind = np.argsort(ar)

    def top_n_union(array,n):
        indices = array[:,-n:]
        indices = np.unique(indices.flatten())
        return indices

    def findcutoff(low,high, lp = -1):
        probe = int((low+high)/2)
        y = top_n_union(ind,probe) # hvg_ids_from_union(adatas,probe)
        if probe == lp:
            return y
        if len(y) > numgenes:
            return findcutoff(low,probe,probe)
        else:
            return findcutoff(probe,high,probe)

    indices = findcutoff(0,numgenes)
    return indices

def subsample_to_min_cellcount(adatas):
        smallest = min([e.X.shape[0] for e in adatas])
        for a in adatas:
            if a.X.shape[0] > smallest:
                sc.pp.subsample(a, n_obs=smallest,
                                random_state=0,
                                copy=False)
        return adatas

import warnings
def subsample(data,num=1000, seed=None, copy = False):
    np.random.seed(seed)
    obs_indices = np.random.choice(data.n_obs, size=num, replace=True)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        r=  data[obs_indices]
        if copy:
            r = r.copy()
        r.obs_names_make_unique()
    return r

def subsample_preprocess(adatas,num = 1000 ,copy = False, **preprocessargs):
    data = Map(subsample,adatas,num=num, copy=copy)
    return preprocess(data,**preprocessargs)



def align(adatas, base = 'pca40' ):
    for i in range(len(adatas)-1):
        hung, _ = hung_adatas(adatas[i], adatas[i+1], base= base)
        adatas[i+1]= adatas[i+1][hung[1]]

def to_array(ad,base):
    return ad.X if not base else ad.obsm[base]

def to_arrays(adatas,base):
    return Map(to_array, adatas, base=base)

def hungarian(adata, adata2, base):
        X = to_arrays([adata, adata2], base=base)
        hung, dist = hung_nparray(*X)
        return hung, dist[hung]

def hung_nparray(X1, X2, debug = False,metric='euclidean'):
    distances = pairwise_distances(X1,X2, metric=metric)
    row_ind, col_ind = linear_sum_assignment(distances)
    if debug:
        x = distances[row_ind, col_ind]
        num_bins = 100
        print("hungarian: debug hist")
        plt.hist(x, num_bins, facecolor='blue', alpha=0.5)
        plt.show()
    return (row_ind, col_ind), distances



def cell_ranger(adatas, mingenes = 200,
                        normrow= True,
                        log = True):
    if 'cell_ranger' in adatas[0].var:
        return adatas

    return Map( lambda x:cell_ranger_single(x, mingenes=mingenes, normrow= normrow,  log= log), adatas)

def cell_ranger_single(adata,
                        mingenes = 200,
                        normrow= True,
                        log = True):

    okgenes = sc.pp.filter_genes(adata, min_counts=3, inplace=False)[0]
    sc.pp.normalize_total(adata, 1e4)
    sc.pp.log1p(adata)
    adata2 = adata[:,okgenes].copy()
    sc.pp.highly_variable_genes(adata2, n_top_genes=5000,
                                         flavor='cell_ranger',
                                        inplace=True)

    fullscores = np.full(adata.X.shape[1],np.NINF,np.float)
    fullscores[okgenes]  = adata2.var['dispersions_norm']
    adata.var['cell_ranger']=  fullscores
    return adata




from lucy import embed

def lapgraph(adatas,base = 'pca40', intra_neigh = 15, inter_neigh = 1,
              scaling_num_neighbors = 2, outlier_threshold = .8,
              scaling_threshold = .9, dataset_adjacency = None):

    X = to_arrays(adatas, base)
    graph =  embed.linear_assignment_integrate(X,
                            intra_neigh=intra_neigh,
                            inter_neigh = inter_neigh,
                            scaling_num_neighbors = scaling_num_neighbors,
                            outlier_threshold = outlier_threshold,
                            scaling_threshold=scaling_threshold,
                                dataset_adjacency=dataset_adjacency)
    return (graph, base)

def graph_embed(adatas, lapgraph, n_components= 2, label = 'lap'):
    # do the embedding
    graph, base = lapgraph
    X = to_arrays(adatas, base)
    projection = embed.distmatrixumap(X,graph, components = n_components)
    adatas = attach_stack(adatas, projection,label)
    return adatas




def pca(adatas, dim=40, label = 'pca40'):

    if label in adatas[0].obsm:
        print('redundant pca :) ')
    # get a result
    data = stack(adatas)
    scaled = sc.pp.scale(data, zero_center=False, copy=True,max_value=10).X
    stackedPCA =  pca_on_scaled_data(scaled, dim)
    return attach_stack(adatas, stackedPCA ,label)

def pca_on_scaled_data(scaled, dim):
    if  not issparse(scaled):
        stackedPCA = decomposition.PCA(n_components  = dim).fit_transform(scaled)
    else:
        stackedPCA = sc.pp._pca._pca_with_sparse(scaled,dim)['X_pca']
    return stackedPCA


def umap(adatas, dim = 10, label = 'umap10', start = 'pca40'):

    if label in adatas[0].obsm:
        print('redundant umap :) ')

    attr = data.obsm.get(start,'')
    X = stack_single_attribute(adatas, attr = attr)
    res = uumap.UMAP(n_components = dim).fit_transform(X)
    return attach_stack(adatas, res ,label)


from sklearn.preprocessing import scale
def project(adatas,start='', **kwargs):

    data = stack_single_attribute(adatas,attr= start)

    if 'pca' in kwargs:
        scaled = scale(data, with_mean= False)
        dim = kwargs['pca']
        data =  pca_on_scaled_data(scaled, dim)
        start = f"pca{dim}"
        attach_stack(adatas, data , start)

    if 'umap' in kwargs:
        dim = kwargs['umap']
        data = uumap.UMAP(n_components = dim).fit_transform(data)
        start = f"umap{dim}"
        attach_stack(adatas, data , start)

    if  'lapumap' in kwargs:
        dim = kwargs['lapumap']
        sim = embed.make_adjacency(similarity(adatas),algo=0,neighbors=10)
        gr = lapgraph(adatas,base=start,dataset_adjacency = sim)
        start = f"lapumap{dim}"
        graph_embed(adatas,gr,n_components = dim, label = start)

    return adatas



# predict function that just diffuses labels
def predict(adatas, tartget_id, labelname= 'predictedlabels'):
    pass





import cellsaw.merge.mergehelpers  as mergehelpers
import numpy as np
from cellsaw.merge.diffusion import Diffusion, stringdiffuse
from cellsaw.draw import plot_merge, confuse2
import logging
class mergeutils:
    def getlabels(self,masked = [], label = 'true', fill = -1):
        f = lambda i,e: np.full_like(e.obs[label],fill)  if i in masked else e.obs[label]
        return [ f(i,e) for i,e in enumerate(self.data)]

    def confuse2(self, labels):
        assert self.sorted
        assert len(self.data) == 2, 'i could just assume that you mean the first 2 datasets, but i refuse :D'
        assert len(labels[0])== self.data[0].shape[0]
        assert len(labels[1])== self.data[1].shape[0]

        confuse2(labels)

    def size(self, pid=-1):
        return [x.shape for x in self.projections[pid]]

    def plot(self, labels, **kwargs):
        plot_merge(self, labels,**kwargs)



class Merge(mergeutils):
    def __init__(self, adatas, selectgenes = 2000,
            make_even = True, pca = 40, umaps = [2],
            joint_space = True,
            sortfield = 0,
            oldcut = False,
            genescoresid = '',
            titles = "ABCDEFGHIJKIJ"):
        assert isinstance(adatas, list), f'merge wants a list, not {type(adatas)}'
        shapesbevorepp= [a.X.shape for a in adatas]
        assert all([a.X.shape[1] == adatas[0].X.shape[1] for a in adatas])
        #self.genescores = [a.varm['scores'] for a in adatas]

        scorename = genescoresid or adatas[0].uns['lastscores']
        self.genescores = [a.varm[scorename] for a in adatas]
        self.similarity = calc_similarity(self.genescores)
        #self.geneab = [a.varm['genes'] for a in adatas]
        if oldcut:
            self.data  = mergehelpers.unioncut(self.genescores, selectgenes, adatas)
        else:
            self.data  = mergehelpers.equal_contrib_gene_cut(self.genescores, selectgenes, adatas)
        self.sorted = False
        self.jointSpace = joint_space

        # geneab = np.all(np.array(self.geneab), axis=0)
        # for i, d in enumerate(self.data):
        #     data[i] = d[:, geneab]

        if make_even:
            self.data = mergehelpers.make_even(self.data)



        logging.info("preprocess:")
        logging.info(f"{make_even=}")
        for a,b in zip(shapesbevorepp, self.data):
            logging.info(f"{a} -> {b.shape}")

        ######################
        self.titles = titles

         # do dimred
        self.projections = [[ d.X for d in self.data]]+mergehelpers.dimension_reduction(self.data,
                                                                False, # scale (will still scale if pca)
                                                                False, # will be ignored anyway
                                                                PCA=pca,
                                                                umaps=umaps,
                                                                joint_space=joint_space)

        if pca:
            self.PCA = self.projections[1]

        if sortfield >=0:
            assert make_even==True, 'sorting uneven datasets will result in an even dataset :)'
            self.sort_cells(sortfield)

        if umaps:
            for x,d in zip(umaps,self.projections[int(pca>0)+1:]):
                self.__dict__[f"d{x}"] = d


    def sort_cells(self,projection_id = 0):
        # loop over data sets
        self.hungdist =[]
        for i in range(len(self.data)-1):
            hung, dist = self.hungarian(projection_id,i,i+1)
            self.hungdist.append(dist)
            #self.data[i+1] = self.data[i+1][hung[1]]
            #self.data[i+1].X = self.data[i+1].X[hung[1]]
            for x in range(len(self.projections)):
                self.projections[x][i+1] = self.projections[x][i+1][hung[1]]
                if x == 0:
                    self.data[i+1]= self.data[i+1][hung[1]]

        self.sorted = True

    #
    def hungarian(self,data_fld,data_id, data_id2):
            hung, dist = mergehelpers.hungarian(self.projections[data_fld][data_id],self.projections[data_fld][data_id2])
            return hung, dist[hung]


from sklearn.metrics import accuracy_score as acc
def accuracy_evaluation(target, true = '', predicted = ''):
    t = target.obs[true]
    p = target.obs[predicted]
    t =list(t)
    p=list(p)
    #for a in zip(t,p): print (a)
    return acc(t,p)



import ubergauss.tools as ut
def calc_similarity(scorelist):
    def dist(a,b):
        asd = np.array([ ut.binarize(d,2000)  for d in [a,b] ])
        union = np.sum(np.any(asd, axis=0))
        intersect = np.sum(np.sum(asd, axis=0) ==2)
        return intersect/union
    res = [[ dist(a,b) for a in scorelist] for b in scorelist]
    return np.array(res)
